##############################################################################
# 
# Template processor library
# --------------------------
#
# Version: 1.0.17        Date: 11/07/2005
#
# Author: Cesar Lobo <cesar.lobo@uam.es>
#
##############################################################################

package NCM::Template;

use strict;

BEGIN {

    require Exporter;
    our @ISA = ("Exporter");
    our @EXPORT = qw(Substitute SetDelimiters);

}

##############################################################################
# Globals
##############################################################################

my $var_list = {};    # Hash with the variables
 
my @file_list = ();   # Array to manage the filenames in the error messages
my $file_index = 0;
 
my $comp_name = '';   # Component name (provided by the user)

my $self = undef;     # pointer to object in use...

my $left='<'; my $right='>';

##############################################################################
#
# ReadFile Subroutine
# -------------------
#
# Arguments:
#
#  1) Name of the file to be readed
#
# Return values:
#
#  The data from the specified file or undef if error
#
##############################################################################

sub ReadFile($) {

    my $file = shift;
    my $FILE; my $buffer;

    local $/ = undef;

    if ( open($FILE,"<$file") ) {
        $buffer = <$FILE>;
        close($FILE);
        return($buffer);
    }

    undef;

}

##############################################################################
#
# Savefile Subroutine
# -------------------
#
# Arguments:
#
#  1) Name of the file
#  2) Data to save
#
# Return values:
#
#  undef value if error
#  1 if ok
#
##############################################################################

sub SaveFile($$) {

    my $file = shift;
    my $data = shift;

    my $FILE;

    return(undef)
        unless( open($FILE,">$file") );

    print $FILE $data;

    return(1);

}

##############################################################################
#
# Lookup_Resource wrapper
# -----------------------
#
# Arguments:
#
#  1) Configuration object
#  2) String with the resource path
#
# Return values:
#
#  Reference to the array with the resource values or undef if error
#
##############################################################################

sub Lookup_Resource($$) {

    my $cfg = shift;
    my $string = shift;

    my $element;

    if($cfg->elementExists($string)) {

        $element = $cfg->getElement($string);

        if($element->isResource()) {
            return(Lookup($element,1));
        }

    }

    return(undef);
   
}

##############################################################################
#
# Lookup_Property wrapper
# -----------------------
#
# Arguments:
#
#  1) Configuration object
#  2) String with the property path
#
# Return values:
#
#  Property value or undef if error
#
##############################################################################

sub Lookup_Property($$) {

    my $cfg = shift;
    my $string = shift;
 
    my $element;

    if($cfg->elementExists($string)) {
    
        $element = $cfg->getElement($string);
    
        if($element->isProperty()) {
            return(Lookup($element,0));
        }
 
    }
 
    return(undef);

}

##############################################################################
#
# Lookup Subroutine
# -----------------
#
# Arguments:
#
#  1) Element path
#  2) Lookup type (1 = Resource Lookup, 0 = Property Lookup)
#
# Return values:
#
#   * A hard reference to the array with all the values (if resource lookup)
#   * The property value (if property lookup) 
#   * undef if error
#
##############################################################################

sub Lookup($$) {

    my $element = shift;
    my $type = shift;

    my $items = [];

    return($element->getValue())
        unless($type == 1);

    while($element->hasNextElement()) {

        my $next = $element->getNextElement();
        push (@$items, $next->getName());
     
    }

    return($items);

}

##############################################################################
# Small helper function for escaping chars
##############################################################################

sub escape($) {

    my $str = shift;

    $str =~ s/([^a-zA-Z0-9])/sprintf("_%lx", ord($1))/eg;

    # special case: 1st character cannot be a digit
    $str =~ s/^([0-9])/sprintf("_%lx", ord($1))/e;

    return($str);

}

##############################################################################
# Small helper function for unescaping chars
##############################################################################

sub unescape($) {

    my $str = shift;

    $str =~ s!(_[0-9a-f]{2})!sprintf("%c",hex($1))!eg;

    return($str);

}


# Disk sorted foreach. Traverses a list of partitions in the order
# they should appear on disk, which is different from the
# lexicographical order panc imposes.
sub disk_sort ($) {
	my $disklist = shift;

	return (sort {my $regexp = '.+[a-z]([0-9]+)';
		     $a =~ m/$regexp/;
		     my $dp1 = $1;
		     $b =~ m/$regexp/;
		     my $dp2 = $1;
		     $dp1<=>$dp2;
	     } @$disklist);
}


##############################################################################
#
# Eval Subroutine
# ---------------
#
# Arguments:
#
#  1) Configuration object
#  2) Data (as generated by Parse)
#
# Return values:
#
#   Data evaluated or undef if error
#
##############################################################################

sub Eval($$) {

    my $cfg = shift;
    my $tokens = shift;

    my $data = '';

    foreach my $token (@$tokens) {
    
        my $type = $token->{TOKEN};

        # TEXT token
        if($type eq 'TEXT') {
            $data .= $token->{DATA};
        }

        # INCLUDE token
        if($type eq 'INCLUDE') {
            my $file; my $d; my $aux;
            $file = &Eval($cfg,$token->{DATA});

            unless(defined($file)) { 
                return(undef); 
            }

            $file =~ s/^\s+//; $file =~ s/\s+$//;

            unless($d=ReadFile($file)) {
                $self->error("in ($file_list[$file_index]) line $token->{LINE}: ".
                             "Unable to open the file $file");
                return(undef);
            }

            # Push the filename in the file list
            push(@file_list,$file);

            $file_index++;

            return(undef)
                unless( defined($aux = Process_Template($cfg,$d)) );

            pop(@file_list); $file_index--;

            $data .= $aux;

        }

        # DATE token
        if($type eq 'DATE') {
            $data .= $token->{DATA};
        }

        # PACKAGE token
        if($type eq 'PACKAGE') {

            my $rpmfile; my $path; my $d; my $ver; my $arch; my $wantarch;

            $rpmfile = &Eval($cfg,$token->{DATA});
            $rpmfile =~ s/^\s+//; $rpmfile =~ s/\s+$//;

            #
            # architecture can be defined with the following syntax:
            #   <packagename>:<path> eg. kernel:i686
            #
            if ($rpmfile =~ m%^(\S+)\:(\w+)$%) {
                $rpmfile = $1;
                $wantarch = $2;
            }

            $path = "/software/packages/" . escape($rpmfile);

            # Get the package version list
            $d = Lookup_Resource($cfg,$path);

            # Substitute only if the package exists
            if(defined($d->[0])) {

                # only the last version is wanted
                $ver = $d->[$#$d];

                # take arch from profile if not already defined
                # For backward compatibility....
                $arch = Lookup_Property($cfg,$path."/$ver/arch");

                if( defined($arch) ) {
                    if (defined $wantarch && $wantarch ne $arch) {
                        $self->error("selected architecture $wantarch not defined in profile");
                        return undef;
                    } else {
                        $data .= $rpmfile."-".unescape($ver).".$arch.rpm";
                    }
                } else {
                    # multiple arch schema. By default, get the first arch present
                    $arch = Lookup_Resource($cfg,$path."/$ver/arch");
                    unless (defined $wantarch) {
                        $arch = Lookup_Property($cfg,$path."/$ver/arch/$arch->[0]");
                    } else {
                        # a specific architecture has been requested. Check if it is in the profile
                        my $found = 0;
                        for (my $i=0; $i <= $#$arch; $i++) {
                            if (Lookup_Property($cfg,$path."/$ver/arch/$arch->[$i]") eq $wantarch) {
                                $found = 1;
                            }
                        }
                        unless ($found) {
                            $self->error("selected architecture $wantarch not defined in profile");
                            return undef;
                        } else {
                            $arch = $wantarch;
                        }
                    }
                    $data .= $rpmfile . "-" . unescape($ver) . ".$arch.rpm";
                }

            } else {
                $self->error("in ($file_list[$file_index]) line $token->{LINE}: ".
                             "The package $rpmfile is not defined in the profile");
                return(undef);
            }
        }

        # PACKAGELIST token
        if ( $type eq 'PACKAGELIST' ) {
            my $pkgname = &Eval($cfg,$token->{DATA});
            $pkgname =~ s/^\s+//; $pkgname =~ s/\s+$//;

            my $wantarch;
            if ($pkgname =~ m%^(\S+)\:(\w+)$%) {
                $pkgname = $1;
                $wantarch = $2;
            }

            my $path = "/software/packages/" . escape($pkgname);

            # Get the package version list
            my $d = Lookup_Resource($cfg,$path);

            foreach my $ver (@$d) {
                # For backward compatibility, first try arch as property
                my $arch = Lookup_Property($cfg,$path."/$ver/arch");

                if ( defined($arch) ) {
                    if (defined($wantarch) && $wantarch ne $arch) {
                        $self->error("selected architecture $wantarch not defined in profile");
                        return undef;
                    }
                }
                else {
                    # Now try arch again, this time as a Resource
                    $arch = Lookup_Resource($cfg,$path."/$ver/arch");
                    if (defined($wantarch)) {
                        my $found = 0;
                        for (my $i=0; $i <=$#$arch; $i++) {
                            if (Lookup_Property($cfg,$path."/$ver/arch/$arch->[$i]") eq $wantarch) {
                                $found = 1;
                                last;
                            }
                            unless ($found) {
                                $self->error("selected architecture $wantarch not defined in profile");
                                return undef;
                            }
                            $arch = $wantarch;
                        }
                    }
                    else {
                        $arch = Lookup_Property($cfg,$path."/$ver/arch/$arch->[0]");
                    }
                }
                if (defined($arch)) {
                    $data .= $pkgname."-".unescape($ver).".".$arch.".rpm ";
                }
            }
        }

        # PACKAGEPATH token
        if($type eq 'PACKAGEPATH') {

            my $rpmfile; my $path; my $d; my $ver; my $rep_name;
            $rpmfile = &Eval($cfg,$token->{DATA});
            $rpmfile =~ s/^\s+//; $rpmfile =~ s/\s+$//;

            my ($rpmversion, $rpmarch);

            # rpmfile is either the name of the rpm
            # or the full package (with version, arch and .rpm suffix)
            if ($rpmfile =~ /(.+)-([^-]+)-([^-]+)\.([^\.]+)\.rpm/) {
                $rpmfile = $1;
                $rpmversion = "$2-$3";
                $rpmarch = $4;
            }

            $path = "/software/packages/".escape($rpmfile);
            my $rep_root = "/software/repositories";

            # Get the package version list
            $d = Lookup_Resource($cfg,$path);

            # Substitute only if the package exists
            if(defined($d->[0])) {
                # if no version was specified, select the last one
                if ($rpmversion) {
                    $ver = escape($rpmversion);
                } else {
                    $ver = $d->[$#$d];
                }

                # Get repository where the package is located
                $rep_name = Lookup_Property($cfg, $path . "/$ver/repository");
                unless ( defined($rep_name) ) {
                    $self->error("in ($file_list[$file_index]) line $token->{LINE}: ".
                                 "Profile inconsistent : repository is missing for package $rpmfile");
                    return(undef);
                }

                # Get repository list
                my $rep_list = Lookup_Resource($cfg,$rep_root);
                unless ( $rep_list ) {
                    $self->error("in ($file_list[$file_index]) line $token->{LINE}: ".
                                 "Profile inconsistent : repository list missing.");
                    return(undef);
                }

                # Get repository URL

                my $rep_url;
                foreach my $rep (@$rep_list) {
                    my $cur_rep_name = Lookup_Property($cfg,$rep_root."/".$rep."/name");
                    # Fix bug #16811: PACKAGEPATH returns wrong repository
                    #                 when using a match operator =~ and there are
                    #                 packages with a similar name (e.g. sl3 and sl3_errata)
                    if ( $cur_rep_name eq $rep_name ) { 
                        my $rep_protocols;
                        $rep_protocols = Lookup_Resource($cfg,$rep_root."/".$rep."/protocols");
                        unless ( $rep_protocols ) {
                            $self->error("in ($file_list[$file_index]) line $token->{LINE}: ".
                                         "Profile inconsistent: protocol list missing for " .
                                         "repository $rep_name.");
                            return(undef);
                        }
                        # Always use first protocol
                        for my $protocol (@$rep_protocols) {
                            $rep_url = Lookup_Property($cfg, $rep_root . "/" . $rep .
                                                             "/protocols/".$protocol."/url");
                            last;
                        }
                        last;
                    }
                }
                unless ( $rep_url ) {
                    $self->error("in ($file_list[$file_index]) line $token->{LINE}: ".
                                 "Unable to find repository url for package $rpmfile.");
                    return(undef);
                }
     
                $data .= $rep_url;

            } else {
                $self->error("in ($file_list[$file_index]) line $token->{LINE}: ".
                             "The package $rpmfile is not defined in the profile");
                return(undef);
            }
        }

        # SHELL token
        if($type eq 'SHELL') {

            my $cmd = &Eval($cfg,$token->{DATA});
            unless(defined($cmd)) {
                return(undef);
            }

            $cmd =~ s/^\s+//; $cmd =~ s/\s+$//;
            my $val = `$cmd`;  $data .= $val;

        }

        # SET variable token
        if($type eq 'SET') {

            my $str = &Eval($cfg,$token->{DATA});
            unless(defined($str)) {
                return(undef);
            }
            $str =~ s/^\s+//; $str =~ s/\s+$//;
            $var_list->{$token->{VAR}} = $str;

        }

        # Reference
        if($type eq 'REF') {

            my $str = &Eval($cfg,$token->{DATA});
            unless(defined($str)) {
                return(undef);
            }
            $str =~ s/^\s+//; $str =~ s/\s+$//;

            # Is the reference a variable?
            if($str =~ /^\$/) {

                my $key;
                my $found = 0;

                foreach $key (keys %$var_list) {
                    if($key eq $str) {
                        $found++; 
                        if( defined($var_list->{$str}) ) {
                            $data .= $var_list->{$str}; 
                        } else {
                            $data .= ''; 
                        }
                        last;
                    }
                }

                unless($found) {
                    $self->error("in ($file_list[$file_index]) line $token->{LINE}: ".
                                 "$str is an undefined variable");
                    return(undef);
                }

            } elsif($str =~ /^\//) {

                my $d = Lookup_Property($cfg,$str);

                if( defined($d) ) { $data .= $d; } else {

                    $self->error("in ($file_list[$file_index]) line $token->{LINE}: ".
                                 "$str is not defined or is not a Property");
                    return(undef);
                }

            } else {

                if($comp_name) {
                    $str = "/software/components/$comp_name/" . $str;
                    my $d = Lookup_Property($cfg,$str);

                    if( defined($d) ) {
                        $data .= $d;
                    } else {
                        self->error("in ($file_list[$file_index]) line $token->{LINE}: ".
                                    "$str is not defined or is not a Property");
                        return(undef);
                    } 

                } else {
                    $self->error("in ($file_list[$file_index]) line $token->{LINE}: ".
                                 "No component name provided");
                    return(undef);
                }
            }
        }

        # IF token
        if($type eq 'IF') {

            my $expr = &Eval($cfg,$token->{EXPR});
            unless(defined($expr)) {
                return(undef);
            }
            $expr =~ s/^\s+//; $expr =~ s/\s+$//;
            if($expr !~ /^\//) {
                if($comp_name) {
                    $expr = "/software/components/$comp_name/" . $expr;
                } else {
                    $self->error("in ($file_list[$file_index]) line $token->{LINE}: ". 
                                 "No component name provided for relative path evaluation");
                    return(undef);
                }
            }

            my $p = Lookup_Property($cfg,$expr);
            if( defined($p) && ($p ne 'false') ) {
                my $d = &Eval($cfg,$token->{BODY});
                if(defined($d)) { $data .= $d; } else { return(undef); }

            } elsif( defined( Lookup_Resource($cfg,$expr) ) ) {

                my $d = &Eval($cfg,$token->{BODY});
                if(defined($d)) { $data .= $d } else { return(undef); }

            } elsif( $expr =~ /^(.*)=\s*"(.*)"$/ ) {

                my ($left,$right) = ($1,$2);
                $left =~ s/\s+$//;

                my $d = Lookup_Property($cfg,$left);
                if( defined($d) && ($d eq $right)) {

                    my $d = &Eval($cfg,$token->{BODY});
                    if(defined($d)) { $data .= $d; } else { return(undef); }

                } elsif($token->{ELSE}) {

                    my $d = &Eval($cfg,$token->{ELSE});
                    if(defined($d)) { $data .= $d; } else { return(undef); }

                }

            } elsif($token->{ELSE}) {
 
                my $d = &Eval($cfg,$token->{ELSE});
                if(defined($d)) { $data .= $d; } else { return(undef); }
 
            } 
        }

        # LIF token
        if($type eq 'LIF') {

            my $expr = &Eval($cfg,$token->{EXPR});
            unless(defined($expr)) {
                return(undef);
            }
            $expr =~ s/^\s+//; $expr =~ s/\s+$//;
            if( $expr =~ /^(.*)=\s*(.*)$/ ) {

                my ($left,$right) = ($1,$2);
                $left =~ s/\s+$//;
                if($left eq $right) {

                    my $d = &Eval($cfg,$token->{BODY});
                    if(defined($d)) { $data .= $d; } else { return(undef); }

                } elsif($token->{ELSE}) {

                    my $d = &Eval($cfg,$token->{ELSE});
                    if(defined($d)) { $data .= $d; } else { return(undef); }

                }

            } else { 
                $self->error("in ($file_list[$file_index]) line $token->{LINE}: ".
                             "Syntax error in LIT token");
                return(undef);
            }
        }

        # FOREACH token
        if ($type eq 'FOREACH') {

            my $expr = &Eval($cfg,$token->{EXPR});
            unless(defined($expr)) {
                return(undef);
            }
            $expr =~ s/^\s+//; $expr =~ s/\s+$//;
            if($expr !~ /^\//) {
                $expr = "/software/components/$comp_name/" . $expr;
            }

            my $iter = Lookup_Resource($cfg,$expr);
            if(defined($iter)) {
                foreach my $item (@$iter) {
                    $var_list->{$token->{VAR}} = $item;
                    my $d = &Eval($cfg,$token->{BODY});
                    if(defined($d)) { $data .= $d } else { return(undef); }
                }
            }
        }

	# DSSTFOREACH
	if ($type eq 'DSSTFOREACH') {

            my $expr = &Eval($cfg,$token->{EXPR});
            unless(defined($expr)) {
                return(undef);
            }
            $expr =~ s/^\s+//; $expr =~ s/\s+$//;
            if($expr !~ /^\//) {
                $expr = "/software/components/$comp_name/" . $expr;
            }

            my $iter = Lookup_Resource($cfg,$expr);
            if(defined($iter)) {
                foreach my $item (disk_sort($iter)) {
                    $var_list->{$token->{VAR}} = $item;
                    my $d = &Eval($cfg,$token->{BODY});
                    if(defined($d)) { $data .= $d } else { return(undef); }
                }
            }
        }

        # LFOREACH token
        if ($type eq 'LFOREACH') {
            my $expr = &Eval($cfg,$token->{EXPR});
            unless(defined($expr)) {
                return(undef);
            }
            $expr =~ s/^\s+//; $expr =~ s/\s+$//;
            foreach my $item (split /\s+/, $expr) {
                $var_list->{$token->{VAR}} = $item;
                my $d = &Eval($cfg,$token->{BODY});
                if(defined($d)) { $data .= $d } else { return(undef); }
            }
        }

        # UNESCAPE token
        if ($type eq 'UNESCAPE') {
            my $str = &Eval($cfg, $token->{DATA});    
            $data .= &unescape($str);
        }

    }

    return($data);

}

##############################################################################
#
# Parse Subroutine
# ----------------
#
# Arguments:
#
#  1) Token list (as generated by Scan)
#
# Return values:
#
#  1) Ref to a new list of tokens (undef = error)
#  2) Type of the last token (undef = no last token)
#  3) Line number of the last token (undef = no last token)
#
##############################################################################

sub Parse($) {

    my $tokens = shift;

    my $newtokens = [];
    while(@$tokens) {
   
        my $token = shift(@$tokens);
        my $type = $token->{TOKEN};

        if($type eq 'TEXT') {
            push (@$newtokens, $token); next; 
        }

        if( ($type eq 'ENDIF') || ($type eq 'ELSE') || 
            ($type eq 'ENDFOR') || ($type eq 'ENDLIF') || ($type eq 'ENDLFOR') ||
	    ($type eq 'ENDDSSTFOR')) {
    
            return($newtokens,$type,$token->{LINE}); 

        }

        if($type eq 'IF') {

            my ($body,$term) = &Parse($tokens);
            if($body) {
 
                $token->{BODY} = $body;

                if($term eq 'ELSE') {
                    ($body,$term) = &Parse($tokens);
                    $token->{ELSE} = $body;
                }

                if($body) {
                    unless( defined($term) && ($term eq 'ENDIF')) { 
                        $self->error("in ($file_list[$file_index]) line $token->{LINE}: ".
                                     "<ENDIF> clause missing");
                        return(undef);
                    }

                    ($body,$term) = &Parse($token->{DATA});
                    $token->{EXPR} = $body;

                } else { return(undef); }

            } else { return(undef); }
    
        }

        if($type eq 'LIF') {

            my ($body,$term) = &Parse($tokens);

            if($body) {

                $token->{BODY} = $body;
                if($term eq 'ELSE') {
                    ($body,$term) = &Parse($tokens);
                    $token->{ELSE} = $body;
                }

                if($body) {
                    unless( defined($term) && ($term eq 'ENDLIF')) {
                        $self->error("in ($file_list[$file_index]) line $token->{LINE}: ".
                                     "<ENDLIF> clause missing");
                        return(undef);
                    }

                    ($body,$term) = &Parse($token->{DATA});
                    $token->{EXPR} = $body;

                } else { return(undef); }

            } else { return(undef); }

        }

        if($type eq 'FOREACH') {

            my ($body,$term) = &Parse($tokens);

            if($body) {

                $token->{BODY} = $body;
                unless( defined($term) && ($term eq 'ENDFOR')) {
                    $self->error("in ($file_list[$file_index]) line $token->{LINE}: ".
                                 "<ENDFOR> clause missing");
                    return(undef);
                }

                ($body,$term) = &Parse($token->{DATA});
                $token->{EXPR} = $body;

            } else { return(undef); }

        }

	if($type eq 'DSSTFOREACH') {

            my ($body,$term) = &Parse($tokens);

            if($body) {

                $token->{BODY} = $body;
                unless( defined($term) && ($term eq 'ENDDSSTFOR')) {
                    $self->error("in ($file_list[$file_index]) line $token->{LINE}: ".
                                 "<ENDDSSTFOR> clause missing");
                    return(undef);
                }

                ($body,$term) = &Parse($token->{DATA});
                $token->{EXPR} = $body;

            } else { return(undef); }

        }

        if($type eq 'LFOREACH') {
            my ($body,$term) = &Parse($tokens);
            if($body) {
                $token->{BODY} = $body;
                unless( defined($term) && ($term eq 'ENDLFOR')) {
                    $self->error("in ($file_list[$file_index]) line $token->{LINE}: ".
                                 "<ENDLFOR> clause missing");
                    return(undef);
                }
                ($body,$term) = &Parse($token->{DATA});
                $token->{EXPR} = $body;
            } else { return(undef); }
        }


        push(@$newtokens, $token);

    }

    return($newtokens);

}

##############################################################################
#
# Scan Subroutine
# ---------------
#
# Arguments:
#
#  1) Data to scan
#  2) Starting line number (optional)
#
# Return values:
#
#  1) Ref to a list of tokens (undef = error)
#  2) Remainder of unscanned data (undef = no data remaining)
#  3) Terminator symbol (undef = EOF)
#  4) Line number of the last line processed.
#
##############################################################################

sub Scan($;$) {

    my $data = shift;
    my $line = shift || 1;

    my $tokens = [];

    while($data =~ /^((.|\n)*?)($left|$right)((.|\n)*)$/) {

        my ($text,$br,$rest) = ($1,$3,$4);

        # Escape char?
        if($text !~ /\\$/) {

            #
            # Clasify tokens...
            #
            if($text) {
                push @$tokens, ({TOKEN=>'TEXT',DATA=>$text,LINE=>$line});
                my $t = $text; $t =~ s/.//g; $line += length($t);
            }

            if($br =~ /^$right$/) {
                return($tokens,$rest,$br,$line);
            }

            if($rest =~ /^\s*IF\s*:\s*((.|\n)*)$/) {
                my ($expr,$more,$term) = &Scan($1,$line);
                unless(defined($term) && $term =~ /^$right$/) {
                    $self->error("in ($file_list[$file_index]) line $line: ".
                                 "missing $right symbol");
                    return(undef);
                }

                # Push the IF token
                push @$tokens, ({TOKEN=>'IF',DATA=>$expr,LINE=>$line});
                $data = $more;

                # Cleaning up
                $data =~ /^((.|\n)*?)$left/;

                if($1 =~ /^(\n+\s*)+$/) {
                    my $t = $1; $t =~ s/.//g; $line += length($t);
                    $data =~ s/^(\n+\s*)+//;
                } else { $data =~ s/^\n//; $line += 1; }

            } elsif($rest =~ /^\s*LIF\s*:\s*((.|\n)*)$/) {

                my ($expr,$more,$term) = &Scan($1,$line);

                unless(defined($term) && $term =~ /^$right$/) {
                        $self->error("in ($file_list[$file_index]) line $line: ".
                                     "missing $right symbol");
                    return(undef);
                }

                # Push the LIF token
                push @$tokens, ({TOKEN=>'LIF',DATA=>$expr,LINE=>$line});
                $data = $more;

                # Cleaning up
                $data =~ /^((.|\n)*?)$left/;

                if($1 =~ /^(\n+\s*)+$/) {
                    my $t = $1; $t =~ s/.//g; $line += length($t);
                    $data =~ s/^(\n+\s*)+//;
                } else { $data =~ s/^\n//; $line += 1; }

            } elsif($rest =~ /^\s*FOREACH\s*:\s*(\$(\w+))\s+((.|\n)*)$/) {

                my $var = $1;

                my ($expr,$more,$term) = &Scan($3,$line);

                unless(defined($term) && $term =~ /^$right$/) {
                    $self->error("in ($file_list[$file_index]) line $line: ".
                                 "missing $right symbol");
                    return(undef);
                }

                # Put a new variable in the variable list (value = undef)
                $var_list->{$var} = undef;

                # Push the FOREACH token
                push @$tokens, ({TOKEN=>'FOREACH',VAR=>$var,DATA=>$expr,LINE=>$line});
                $data = $more;

                # Cleaning up
                $data =~ /^((.|\n)*?)$left/;

                if($1 =~ /^(\n+\s*)+$/) {
                    my $t = $1; $t =~ s/.//g; $line += length($t);
                    $data =~ s/^(\n+\s*)+//; 
                } else { $data =~ s/^\n//; $line += 1; }

            }elsif($rest =~ /^\s*DSSTFOREACH\s*:\s*(\$(\w+))\s+((.|\n)*)$/) {

                my $var = $1;

                my ($expr,$more,$term) = &Scan($3,$line);

                unless(defined($term) && $term =~ /^$right$/) {
                    $self->error("in ($file_list[$file_index]) line $line: ".
                                 "missing $right symbol");
                    return(undef);
                }

                # Put a new variable in the variable list (value = undef)
                $var_list->{$var} = undef;

                # Push the FOREACH token
                push @$tokens, ({TOKEN=>'DSSTFOREACH',VAR=>$var,DATA=>$expr,LINE=>$line});
                $data = $more;

                # Cleaning up
                $data =~ /^((.|\n)*?)$left/;

                if($1 =~ /^(\n+\s*)+$/) {
                    my $t = $1; $t =~ s/.//g; $line += length($t);
                    $data =~ s/^(\n+\s*)+//; 
                } else { $data =~ s/^\n//; $line += 1; }

            } elsif($rest =~ /^\s*LFOREACH\s*:\s*(\$(\w+))\s+((.|\n)*)$/) {

                my $var = $1;
                my ($expr,$more,$term) = &Scan($3,$line);

                unless(defined($term) && $term =~ /^$right$/) {
                    $self->error("in ($file_list[$file_index]) line $line: ".
                                 "missing $right symbol");
                    return(undef);
                }

                # Put a new variable in the variable list (value = undef)
                $var_list->{$var} = undef;

                # Push the LFOREACH token
                push @$tokens, ({TOKEN=>'LFOREACH',VAR=>$var,DATA=>$expr,LINE=>$line});
                $data = $more;

                # Cleaning up
                $data =~ /^((.|\n)*?)$left/;

                if($1 =~ /^(\n+\s*)+$/) {
                    my $t = $1; $t =~ s/.//g; $line += length($t);
                    $data =~ s/^(\n+\s*)+//; 
                } else { $data =~ s/^\n//; $line += 1; }

            } elsif($rest =~ /^\s*SET\s*:\s*(\$(\w+))\s*=\s*((.|\n)*)$/) {

                my $var_name = $1;

                my ($expr,$more,$term) = &Scan($3,$line);

                unless(defined($term) && $term =~ /^$right$/) {
                    $self->error("in ($file_list[$file_index]) line $line: ".
                                 "missing $right symbol");
                    return(undef);
                }

                # Create the new variable
                $var_list->{$var_name} = undef;

                # And push the SET token
                push @$tokens, ({TOKEN=>'SET',VAR=>$var_name,DATA=>$expr,LINE=>$line});
                $data = $more;

            # Cleaning up
                if($data =~ /^((.|\n)*?)$left/) {

                    if($1 =~ /^(\n+\s*)+$/) {
                        my $t = $1; $t =~ s/.//g; $line += length($t);
                        $data =~ s/^(\n+\s*)+//;
                    } else { $data =~ s/^\n//; $line += 1; }

                } else {
                    $data =~ /^((.|\n)*?)$/;
                    if($1 !~ /^(\n+\s*)+$/) { $data =~ s/^\n//; $line += 1; }
                }

            } elsif($rest =~ /^\s*SHELL\s*:\s*((.|\n)*)$/) {

                my ($expr,$more,$term) = &Scan($1,$line);

                unless(defined($term) && $term =~ /^$right$/) {
                    $self->error("in ($file_list[$file_index]) line $line: ".
                                 "missing $right symbol");
                    return(undef);
                }

                # Push the SHELL token
                push @$tokens, ({TOKEN=>'SHELL',DATA=>$expr,LINE=>$line});
                $data = $more;

                # Cleaning up
                $data =~ s/^\n//; $line += 1;

            } elsif($rest =~ /^\s*INCLUDE\s*:\s*((.|\n)*)$/) {

                my ($expr,$more,$term) = &Scan($1,$line);

                unless(defined($term) && $term =~ /^$right$/) {
                    $self->error("in ($file_list[$file_index]) line $line: ".
                                 "missing $right symbol");
                    return(undef);
                }

                # Push the INCLUDE token
                push @$tokens, ({TOKEN=>'INCLUDE',DATA=>$expr,LINE=>$line});
                $data = $more;

            } elsif($rest =~ /^\s*PACKAGE\s*:\s*((.|\n)*)$/) {

                my ($expr,$more,$term) = &Scan($1,$line);

                unless(defined($term) && $term =~ /^$right$/) {
                    $self->error("in ($file_list[$file_index]) line $line: ".
                                 "missing $right symbol");
                    return(undef);
                }

                push @$tokens, ({TOKEN=>'PACKAGE',DATA=>$expr,LINE=>$line});
                $data=$more;

            } elsif($rest =~ /^\s*PACKAGELIST\s*:\s*((.|\n)*)$/) {

                my ($expr,$more,$term) = &Scan($1,$line);

                unless(defined($term) && $term =~ /^$right$/) {
                    $self->error("in ($file_list[$file_index]) line $line: ".
                                 "missing $right symbol");
                    return(undef);
                }

                push @$tokens, ({TOKEN=>'PACKAGELIST',DATA=>$expr,LINE=>$line});
                $data=$more;

            } elsif($rest =~ /^\s*PACKAGEPATH\s*:\s*((.|\n)*)$/) {

                my ($expr,$more,$term) = &Scan($1,$line);

                unless(defined($term) && $term =~ /^$right$/) {
                    $self->error("in ($file_list[$file_index]) line $line: ".
                                 "missing $right symbol");
                    return(undef);
                }

                push @$tokens, ({TOKEN=>'PACKAGEPATH',DATA=>$expr,LINE=>$line});
                $data=$more;

            } elsif($rest =~ /^\s*ELSE\s*:\s*$right((.|\n)*)$/) {

                push @$tokens, ({TOKEN=>'ELSE',LINE=>$line});
                $data = $1;

                # Cleaning up
                if($data =~ /^((.|\n)*?)$left/) {

                    if($1 =~ /^(\n+\s*)+$/) {
                        my $t = $1;
                        $t =~ s/.//g;
                        $line += length($t);
                        $data =~ s/^(\n+\s*)+//;
                    } else { $data =~ s/^\n//; $line += 1; }

                } else {

                    $data =~ /^((.|\n)*?)$/;
                    if($1 !~ /^(\n+\s*)+$/) { $data =~ s/^\n//; $line += 1; }

                }

            } elsif($rest =~ /^\s*ENDIF\s*$right((.|\n)*)$/) {

                push @$tokens, ({TOKEN=>'ENDIF',LINE=>$line});
                $data = $1;

                # Cleaning up
                if($data =~ /^((.|\n)*?)$left/) {
                    if($1 =~ /^(\n+\s*)+$/) {
                        my $t = $1; $t =~ s/.//g; $line += length($t);
                        $data =~ s/^(\n+\s*)+//;
                    } else { $data =~ s/^\n//; $line += 1; }
                } else {
                    $data =~ /^((.|\n)*?)$/;
                    if($1 !~ /^(\n+\s*)+$/) { $data =~ s/^\n//; $line += 1; }
                }

            } elsif($rest =~ /^\s*ENDLIF\s*$right((.|\n)*)$/) {

                push @$tokens, ({TOKEN=>'ENDLIF',LINE=>$line});
                $data = $1;

                # Cleaning up
                if($data =~ /^((.|\n)*?)$left/) {
                    if($1 =~ /^(\n+\s*)+$/) {
                        my $t = $1; $t =~ s/.//g; $line += length($t);
                        $data =~ s/^(\n+\s*)+//;
                    } else { $data =~ s/^\n//; $line += 1; }
                } else {
                    $data =~ /^((.|\n)*?)$/;
                    if($1 !~ /^(\n+\s*)+$/) { $data =~ s/^\n//; $line += 1; }
                }

            } elsif($rest =~ /^\s*ENDFOR\s*$right((.|\n)*)$/) { 

                push @$tokens, ({TOKEN=>'ENDFOR',LINE=>$line});
                $data = $1;

                # Cleaning up
                if($data =~ /^((.|\n)*?)$left/) {
                    if($1 =~ /^(\n+\s*)+$/) {
                        my $t = $1; $t =~ s/.//g; $line += length($t);
                        $data =~ s/^(\n+\s*)+//;
                    } else { $data =~ s/^\n//; $line += 1; }
                } else {
                    $data =~ /^((.|\n)*?)$/;
                    if($1 !~ /^(\n+\s*)+$/) { $data =~ s/^\n//; $line += 1; }
                }

            }elsif($rest =~ /^\s*ENDDSSTFOR\s*$right((.|\n)*)$/) { 

                push @$tokens, ({TOKEN=>'ENDDSSTFOR',LINE=>$line});
                $data = $1;

                # Cleaning up
                if($data =~ /^((.|\n)*?)$left/) {
                    if($1 =~ /^(\n+\s*)+$/) {
                        my $t = $1; $t =~ s/.//g; $line += length($t);
                        $data =~ s/^(\n+\s*)+//;
                    } else { $data =~ s/^\n//; $line += 1; }
                } else {
                    $data =~ /^((.|\n)*?)$/;
                    if($1 !~ /^(\n+\s*)+$/) { $data =~ s/^\n//; $line += 1; }
                }

            } elsif($rest =~ /^\s*ENDLFOR\s*$right((.|\n)*)$/) { 
                push @$tokens, ({TOKEN=>'ENDLFOR',LINE=>$line});
                $data = $1;

                # Cleaning up
                if($data =~ /^((.|\n)*?)$left/) {
                    if($1 =~ /^(\n+\s*)+$/) {
                        my $t = $1; $t =~ s/.//g; $line += length($t);
                        $data =~ s/^(\n+\s*)+//;
                    } else { $data =~ s/^\n//; $line += 1; }
                } else {
                    $data =~ /^((.|\n)*?)$/;
                    if($1 !~ /^(\n+\s*)+$/) { $data =~ s/^\n//; $line += 1; }
                }

            } elsif($rest =~ /^\s*DATE\s*$right((.|\n)*)$/) {

                my $date = `date`;
                push @$tokens, ({TOKEN=>'DATE',DATA=>$date,LINE=>$line});
                $data = $1;

                # Cleaning up
                $data =~ s/^\n//; $line += 1;

            } elsif($rest =~ /^\s*UNESCAPE\s*:\s*((.|\n)*)$/) {

                my ($expr,$more,$term) = &Scan($1,$line);
                unless(defined($term) && $term =~ /^$right$/) {
                    $self->error("in ($file_list[$file_index]) line $line: ".
                                 "missing $right symbol");
                    return(undef);
                }

                push @$tokens, ({TOKEN=>'UNESCAPE',DATA=>$expr,LINE=>$line});
                $data=$more;

            } else {

                my ($expr,$more,$term) = &Scan($rest,$line);
                unless(defined($term) && $term =~ /^$right$/) {
                    $self->error("in ($file_list[$file_index]) line $line: ".
                                 "missing $right symbol");
                    return(undef);
                }

                # Push the REF token
                push @$tokens, ({TOKEN=>'REF',DATA=>$expr,LINE=>$line});
                $data=$more;

            }

        } else {

            $text =~ s/\\$//;
            $text .= $br;

            push @$tokens, ({TOKEN=>'TEXT',DATA=>$text,LINE=>$line});
            my $t = $text; $t =~ s/.//g; $line += length($t);

            $data=$rest;

        }

    }

    push @$tokens, ({TOKEN=>'TEXT',DATA=>$data,LINE=>$line}) 
        if($data && ($data ne "\n"));

    return($tokens,undef,undef,$line);

}

##############################################################################
#
# Process_Template Subroutine
# ---------------------------
#
# Arguments:
#
#  1) Configuration object (NVA-API)
#  2) Variable with the Template data.
#
# Returns:
#
#  The final processed data or undef in other case
#
##############################################################################

sub Process_Template($$) {

    my $cfg = shift;
    my $data = shift;

    my $tokens; my @aux; my $output;

    ##################################################
    # Scan
    ##################################################

    ($tokens, @aux) = Scan($data);

    return(undef)
        unless(defined($tokens));

    if(defined($aux[1])) {
        $self->error("in ($file_list[$file_index]) line $aux[2]: ".
                     "unmatched $right found");
        return(undef);
    }

    ##################################################
    # Parse
    ##################################################

    ($tokens, @aux)=Parse($tokens);

    return(undef)
        unless(defined($tokens));

    if(defined($aux[0])) {

        $self->error("in ($file_list[$file_index]) line $aux[1]: ".
                     "unmatched $aux[0] found");
        return(undef);

    }

    ##################################################
    # Eval
    ##################################################

    return(undef)
        unless( defined($output=Eval($cfg,$tokens)) );

    ##################################################
    # End of the template processing
    ##################################################

    return($output);

}

##############################################################################
#
# SetDelimiters Subroutine
# ------------------------
#
# Arguments:
#
#  1) The new value for the left delimiter
#  2) The new value for the right delimiter
#
# Returns:
#
#  1 if the delimiter values were changed
#  undef in other case
#
##############################################################################
                                                                                                                             
sub SetDelimiters($$$) {

    $self=shift;
    if( defined($_[0]) && defined($_[1]) ) {
        $left = shift; $right = shift;
        ( $left ne $right ) ? return(1) : return(undef);
    }

    undef;

}

##############################################################################
#
# GetDelimiters Subroutine
# ------------------------
#
# Arguments:
#
# none
#
# Returns:
#
# an array with the left and the right delimiters 
#
##############################################################################
                                                                                                                             
sub GetDelimiters($) {

    $self=shift;
    return ($left,$right);

}



##############################################################################
#
# Substitute Subroutine
# ---------------------
#
# Input arguments:
#
#  1) Configuration object (NVA-API)
#  2) output filename
#  3) Component name
#  4) Template filename (optional)
#
# Returns:
#
#  1 if the file has been changed
#  undef if error
#
##############################################################################

sub Substitute($$$$;$) {
  
    $self=shift; # this is the actual entry point
    my ($cfg,$output,$comp,$template) = @_;
  
    my $data; my $outfile;

    $comp_name = $comp;

    # Check if a component name has been provided
    unless($template) {
        if($comp_name) {
            unless( $template = Lookup_Property($cfg,"/software/components/$comp_name/template") ) {
                $self->error("/software/components/$comp_name/template is not defined");
                return(undef);
            }
        } else {
            $self->error("A template file or component name must be provided");
            return(undef);
        }
    }
  
    if(not $data=ReadFile($template)) {
        $self->error("Unable to open the file $template");
        return(undef);
    }

    push(@file_list,$template);

    # Process the Template
    return(undef)
        unless( defined($outfile=Process_Template($cfg,$data)) );

    # and save the output
    unless(SaveFile($output, $outfile)) {
        $self->error("Unable to create the output file $output");
        return(undef);
    }

    return(1);

}

1;

##############################################################################

=pod
                                                                                                                             
=head1 NAME

The Quattor Template Processor

=head1 VERSION

1.0.17

=head1 DESCRIPTION

The Quattor Template Processor is a library for processing template files which are 
based on a tag language provided by the library. This tag language supports profile 
properties substitution, conditionals based on profile properties values, iteration 
based on profile resources and other functionality (see below). The library can be 
used to create configuration files very easily, with almost no additional code.

The Quattor Template Processor provides a command line utility called ncm-template-test 
intended for testing purposes.

=head1 FUNCTIONS PROVIDED BY THE LIBRARY

The library has to be included in the inheritance of the module using it:

       use NCM::Template;
       use vars qw(@ISA);

       @ISA= qw(.... NCM::Template);

=over 4

=item B<$result = $self-E<gt>Substitute($cfg, $output, $component [, $template])>

=item Input:

$cfg       -> Configuration object with profile (see NVA-API)

$output    -> Output filename

$component -> Component name:

              The template name will be taken from the path:

                /software/components/component-name/template

          and all the template processing will be done in the 
          context of the provided component. If no component 
          name is needed for the template processing an undef 
          value must be given as argument.

$template  -> Template filename (optional argument) 

              If no component name is specified ($component = undef), 
          this will be the template file to be processed.

=item Output:

$result    -> 1 if the file has been changed 
              undef if error

=item B<$result = $self-E<gt>SetDelimiters($left, $right)>

=item Input:

$left      -> Left delimiter
              default value: '<'

$right     -> Right delimiter
              default value: '>'

=item Output:

$result    -> 1 if delimiters were changed successfuly 
              undef if error

=back

=head1 LANGUAGE SPECIFICATIONS

All the template tags should be enclosed by delimiters. Delimiters are configurable and they can 
be escaped with a backslash, for example, '\<' for the default left delimiter and '\>' for the 
right one. All the text in the template file not enclosed with delimiters is treated as plain
text and will be added to the final output untouched.

=over 4

=item B<E<lt>/path/to/propertyE<gt> (absolute path to property)>

Substitute by the value of the named property. Note that if the path is not defined or
is not a property an error will be given.

=item B<E<lt>relative/path/to/propertyE<gt> (relative path to property)>

The path is relative to the component definition (/software/component/component_name/). If no 
component is given to the Substitute function (or by the ncm-template-test --component option) 
an error will be given.

=item B<E<lt>IF: /path/to/elementE<gt> text E<lt>ELSE:E<gt> text E<lt>ENDIFE<gt>>

If the element (either resource or property) is defined, or is a boolean property and its value is 
true, then substitute the first text, otherwise substitute the second one. The else part is optional.

=item B<E<lt>IF: /path/to/property = "string value"E<gt> text E<lt>ELSE:E<gt> text E<lt>ENDIFE<gt>>

If the element specified is a property and its value is equal to the string value (string equality), 
then substitute the first text, otherwise substitute the second one. The else part is optional.

=item B<E<lt>LIF: string1 = string2E<gt> text E<lt>ELSE:E<gt> text E<lt>ENDLIFE<gt>>

Literal IF. Performs a literal string comparison. If string1 is equal to string2 then the first text is 
substituted, otherwise is substituted the second one. Notice that is a literal string comparision because 
the strings are not interpreted as paths to profile elements, their literal values are the ones involved 
in the comparision. With this tag is posible to check the value of a previous defined variable (<$var> tag) 
or make comparisons between two of them. The else part is optional. 

=item B<E<lt>FOREACH: $var /path/to/resourceE<gt> text E<lt>ENDFORE<gt>>

Substitute one copy of the specified text for each item in the resource (either a list or a 
named list). Inside the loop the variable can be referenced by enclosing with the delimiters, 
for example E<lt>$varE<gt>. If the path specified is not a resource or does not exist, the 
FOREACH block is not evaluated.

=item B<E<lt>DSSTFOREACH: $var /path/to/resourceE<gt> text E<lt>ENDDSSTFOR<gt>

DiSk SorTed FOREACH. Similar to FOREACH, but substitution is sorted
for the partitions to be processed in the same order they will appear
on disk (sda5 before sda10) instead of the lexicographical order used
on the profiles (sda10 before sda5).

=item B<E<lt>LFOREACH: $var value1 value2 ... E<gt> text E<lt>ENDLFORE<gt>>

Literal FOREACH.
Substitute one copy of the specified text for each of the values. 
Inside the loop the variable can be referenced by enclosing with the delimiters,
for example E<lt>$varE<gt>.

=item B<E<lt>PACKAGE: packageE<gt> or E<lt>PACKAGE: package:archE<gt>>

Replace by package.VER.ARCH.rpm, where VER and ARCH are given by the
profile. If more than one version of the same package is defined in
the profile, the latest defined version will be used. If multiple
architectures are present for the package, the first one will be used,
unless the architecture is defined using the following syntax:
package:arch e.g. kernel:i686.

=item B<E<lt>PACKAGEPATH: packageE<gt>>

Replace by path (url) where the package is located. Algorithm to select
the package is the same as with clause PACKAGE.

Instead of just the package name, also the full package 
(including version, architecture and .rpm suffix)
may be specified.

=item B<E<lt>PACKAGELIST: packageE<gt>>

Return a space-separated list of all packages with the specified package name.
The package is formatted as 
<name>-<version>-<release>.<architecture>.rpm

=item B<E<lt>INCLUDE: filenameE<gt>>

The file specified is parsed and the result added to the final output.

=item B<E<lt>SET: $var = exprE<gt>>

Set a global variable to the given value. The variable can be referenced by enclosig with the 
delimiters, for example E<lt>$varE<gt>

=item B<E<lt>SHELL: expresionE<gt>>

Executes the given expresion and the output of the execution (if any) is added to the output 
file.

=item B<E<lt>DATEE<gt>>

Replace with current date as given by the 'date' user command.

=back

=head1 EXAMPLES

See under the /usr/share/doc/ncm-template-1.0.17 directory.

=head1 SEE ALSO

ncm-template-test

=head1 AUTHOR

Cesar Lobo <cesar.lobo@uam.es>

=cut

##############################################################################
